/*********************************************
Project : Heizung ATmega32
Version : 2.0
Date    : 09.12.2003  // AT90C8535
Date    : 08.01.2006  // ATMEGA32
Date    : 29.11.2006  // Fehler im Taktgenerator (Uhr ging falsch)
Author  : (c) Holger Buss
http://www.mikrocontroller.com --> Bauanleitungen
Chip type           : ATmega32
Program type        : Application
Clock frequency     : 8,000000 MHz
Memory model        : Small
External SRAM size  : 0
Data Stack size     : 512
*********************************************/
#define EXTRA_MENU 
//#define INCR_GEBER 
//#define NO_LCD


#include <mega32.h>
// Alphanumeric LCD Module functions
#include <delay.h>
#include <stdio.h>

#ifdef NO_LCD
#include "lcd_remote.h"                         
#include "lcd_remote.c"                         
#else
// Alphanumeric LCD Module functions
#include <lcd.h>                         
#asm
   .equ __lcd_port=0x15
#endasm
#endif

//-----------
typedef eeprom struct
{
  unsigned int Zeit;
  unsigned char Mode;
} struct_daten;

eeprom struct_daten Timer[65];
eeprom unsigned char SchichtOffset;
unsigned int AktuellZeit = 0;
bit Polling100ms = 1;
#define WERKTAG       0x4000
#define ZEIT(s,m)     (s * 60 + m)
#define ZEIT_MASKE    7*1440
flash char Tag[7,3]= {"Mo","Di","Mi","Do","Fr","Sa","So"};
#define NACHT         0
#define TAG           1
#define ABWESEND      2
#define DEAKTIV       3
unsigned char BetriebsMode = TAG;
unsigned int CountHeizzeit = 0,CountZeit = 0;

flash unsigned char Leds[3] = { 0x04, 0x08, 0x10 };
//-----------
unsigned char Taste = 0, MenuCount=0;
bit MenuPolling = 1;
bit RefreshDisplay = 0;
bit PollingMinute = 0;                           
bit PollingTag = 0;
bit SendeDaten = 0;
bit Output = 0;
bit SendeDisplay = 0;

#define NICHTS 0
#define MENU   1
#define UP     3
#define DOWN   4
#define OKAY   5
#define MODE   6

#define TASTE1       PINA.7
#define TASTE2       PINA.6
#define TASTE3       PINA.5
#define TASTE4       PINA.4
#define TASTE5       PINA.3

//-----------
#define ANZAHL_KENNLINIEN 3

signed int  AussenTemperatur;
signed int  KesselTemperatur;
signed long DurchschnittsTemperatur = 0;        
#define ANZ_LOGGING 250
eeprom unsigned char VerbrauchsRingSpeicher[ANZ_LOGGING];
eeprom unsigned char TemperaturRingSpeicher[ANZ_LOGGING];
eeprom unsigned char ZeigerRingspeicher;

#define LED_NACHT      PORTB.0
#define LED_ABWESEND   PORTB.1
#define RELAIS_KESSEL  PORTB.2
#define RELAIS_PUMPE   PORTB.3
#define RELAIS_TAG     PORTB.4

flash signed int   KesselHysterese = 15;                         // 1,5 Grad Schalthysterese
//#define            KESSEL_REST_TEMPERATUR  400                   // Die Pumpe blebt an, bis diese Temperatur unterschritten ist

signed int          SollKesselTemperatur = 0; 
eeprom signed int  MaxKesselTemperatur[ANZAHL_KENNLINIEN];
eeprom signed int  MinKesselTemperatur[ANZAHL_KENNLINIEN];
eeprom signed int  KennlinieTemperaturOben[ANZAHL_KENNLINIEN];
eeprom signed int  KennlinieTemperaturUnten[ANZAHL_KENNLINIEN];
eeprom signed int  PumpeEinTemperatur[ANZAHL_KENNLINIEN];
eeprom unsigned char CheckEEpromData;
#define EEPROM_KOMPATIBILITAET  1
//-----------
#define GRAD_ZEICHEN  0xDF

#ifndef NO_LCD
flash char lcd_char1[8]=
 {
  0b0000100,
  0b0010101,
  0b0001110,
  0b0010001,
  0b0010001,
  0b0001110,
  0b0010101,
  0b0000100
  };
flash char lcd_char0[8]=
 {
  0b0000000,
  0b0000000,
  0b0001110,
  0b0011111,
  0b0011111,
  0b0001110,
  0b0000000,
  0b0000000
  };
flash char lcd_char2[8]=
 {
  0b0000000,
  0b0000000,
  0b0001110,
  0b0010001,
  0b0010001,
  0b0001110,
  0b0000000,
  0b0000000
  };

flash char lcd_char3[8]=
 {
  0b0000000,
  0b0000100,
  0b0001110,
  0b0010101,
  0b0000100,
  0b0000100,
  0b0000100,
  0b0000000
  };

flash char lcd_char4[8]=
 {
  0b0000000,
  0b0000100,
  0b0000100,
  0b0000100,
  0b0010101,
  0b0001110,
  0b0000100,
  0b0000000
  };



#define RXB8 1
#define TXB8 0
#define UPE 2
#define OVR 3
#define FE 4
#define UDRE 5
#define RXC 7





// ---------------------------------------------------------------------------------------
void LCD_UserChar(char flash *pc,char char_code)
{
  char i,a;
  a=(char_code<<3) | 0x40;   
  for (i=0; i<8; i++) lcd_write_byte(a++,*pc++);
}

#endif


// ---------------------------------------------------------------------------------------
void LCDInit(void)
{
// LCD module initialization
  lcd_init(16);
#ifndef NO_LCD
  LCD_UserChar(lcd_char0,1);
  LCD_UserChar(lcd_char1,2);
  LCD_UserChar(lcd_char2,3);
  LCD_UserChar(lcd_char3,4);
  LCD_UserChar(lcd_char4,5);
#endif
}
// ---------------------------------------------------------------------------------------
#define LCD      0
#define SERIELL  1
#define lcd_int(w, a, p)     Raw_INT(LCD, w, a, p)
#define Seriell_int(w, a, p) Raw_INT(SERIELL, w, a, p)
#define _ALTERNATE_PUTCHAR_
void putchar(char c)
{
  if (Output == LCD) lcd_putchar(c);
   else 
   {
    #asm
     sbis usr,udre
     rjmp _putchar
     ld   r30,y
     out  udr,r30
#endasm
   }
}                                              

flash unsigned int disp_wert[] = {0,1,10,100,0xFF};
#define LEADING_ZERO  -1 
// ---------------------------------------------------------------------------------------
void Raw_INT(char output,signed int wert , signed char anzahl , signed char punkt_pos)
 {
  signed char i;
  unsigned char pos, null;  
   
  Output = output;
  
  if(anzahl < 0)                         
   {
     anzahl = -anzahl;
     if((signed int)wert < 0) { wert = -wert; putchar('-');  }
     else  putchar('+');                    
   }

  if((punkt_pos > anzahl) || (anzahl > 3))
   {
    for (i=0;i<anzahl;i++) putchar('?'); return;                            
   }

  if(wert >= disp_wert[anzahl + 1])     
   {
    for (i=0;i<anzahl;i++) putchar('?');  return;                            
   }

  if(punkt_pos < 0 ) null = 1; else null = 0;
  
  for(i = anzahl ; i > 0 ; i--)
   {
    if(i == punkt_pos) {  putchar('.');  null = 1; } 

    pos = (wert / disp_wert[i]) + 0x30;

    if((pos == '0') && !null && (i != 1))    
      if (i== punkt_pos+1) putchar('0');
        else putchar(' ');
    else
     {
      null = 1;                        
      putchar(pos);
      wert %= disp_wert[i];            
     }
   }   
    
 }
// ---------------------------------------------------------------------------------------
unsigned int read_adc(unsigned char adc_input)
{
 ADMUX=adc_input;
 ADCSR.6=1;
 while (ADCSR.4==0);
 ADCSR.4=1;
 return ((ADCW*25)/4)-2250 ;
}

// ---------------------------------------------------------------------------------------
void GetTemperatur(unsigned char Start)
{
 signed int tmp_int;
 
 tmp_int = read_adc(1);

 if(Start) AussenTemperatur = tmp_int;
 if(PollingMinute) 
  {
    if (AussenTemperatur > tmp_int)      AussenTemperatur--;
    else if (AussenTemperatur < tmp_int) AussenTemperatur++;
    PollingMinute = 0;
  }
 tmp_int = read_adc(2);

 if (KesselTemperatur > tmp_int)      KesselTemperatur--;
  else if (KesselTemperatur < tmp_int) KesselTemperatur++;
 if(Start) KesselTemperatur = tmp_int;
}

// ---------------------------------------------------------------------------------------
// 1ms - Interrupt
interrupt [TIM2_COMP] void timer2_comp_isr(void)
{
  static unsigned int clock= 0,tim = 0;
  
  if (++clock == 60000) 
   { AktuellZeit++; 
     clock = 0; 
     PollingMinute = 1;
     DurchschnittsTemperatur += AussenTemperatur / 10;

     if (AktuellZeit % 1440 == 0) 
      { 
           PollingTag = 1;
      }

     if (AktuellZeit >= 1440*7) 
      { 
       AktuellZeit = 0;
      }
 
   } 
  if(++tim == 2000)
   {    
    if(RELAIS_KESSEL == 0) CountHeizzeit++;
    CountZeit++;
    tim = 0;
   }
  if ((clock % 350) == 0) RefreshDisplay = 1;
  if ((clock % 10 ) == 0) Polling100ms = 1;
}
// ---------------------------------------------------------------------------------------
interrupt [USART_TXC] void uart_tx_isr(void)
{
}

interrupt [USART_RXC] void uart_rx_isr(void)
{
  unsigned char zeichen;
  zeichen = UDR;
  
  switch (zeichen)
  { case 'T':  SendeDaten = 1; break;
    case '2':  Taste = UP; SendeDisplay = 1;break;
    case '1':  Taste = DOWN; SendeDisplay = 1; break;
    case '3':  Taste = MODE; break;
    case ' ':  SendeDisplay = 1; break;
    case '\r': Taste = OKAY; SendeDisplay = 1; break; 
  } 
}

// ---------------------------------------------------------------------------------------
// External Interrupt 0 service routine
#ifdef INCR_GEBER
// External Interrupt 1 service routine
interrupt [EXT_INT1] void ext_int1_isr(void)
{
  unsigned char Taste2;
  
  Taste2 = TASTE2;
  
  if (MCUCR==0x08) // falling edge 
  {
    if (!Taste2) Taste = DOWN; else Taste = UP;
    MCUCR=0x0C;
  }
  else 
  {  MCUCR=0x08;  // raising edge
     if (!Taste2) Taste = UP; else Taste = DOWN;
  }
  GIFR |= 0x80;
}                      
#endif
// ---------------------------------------------------------------------------------------
#define CNT_TASTE 7

void Tastatur(void)
{
 static char taste2 = 0,taste3 = 0,taste4 = 0,taste5 = 0;
 Taste = NICHTS;

#ifdef INCR_GEBER

#else
// if(TASTE1) { if(taste1++ == 0 || taste1 == CNT_TASTE) Taste = MENU;  if(taste1 == CNT_TASTE) taste1 = 3;} else taste1 = 0;
 if(TASTE2) { if(taste2++ == 0 || taste2 == CNT_TASTE) Taste = DOWN;    if(taste2 == CNT_TASTE) taste2 = 3;} else taste2 = 0;
 if(TASTE3) { if(taste3++ == 0 || taste3 == CNT_TASTE) Taste = UP;  if(taste3 == CNT_TASTE) taste3 = 3;} else taste3 = 0;
#endif

 if(TASTE4) { if(taste4++ == 0) Taste = OKAY;} else taste4 = 0;
 if(TASTE5) { if(taste5++ == 0) Taste = MODE;} else taste5 = 0;
}                   
  
// ---------------------------------------------------------------------------------------
unsigned char ErmittleBetriebsMode(unsigned int zeit)
{
   unsigned char rueck = DEAKTIV, i;
   
   for (i = 0; i < 32; i++) 
   {  
      if (Timer[i+ SchichtOffset].Zeit & WERKTAG)                         // Sonderfall Werktags
      {  if (zeit  < (6*1440))                                          // akt. Tag ist Mo - Fr
         if (((Timer[i+ SchichtOffset].Zeit & ~WERKTAG) % 1440)== (zeit % 1440)) { rueck = Timer[i+ SchichtOffset].Mode; break;} // testen  
      }
      else 
      if (Timer[i+ SchichtOffset].Zeit == zeit) { rueck = Timer[i+ SchichtOffset].Mode; break; } 
   }
   return(rueck);
}

// ---------------------------------------------------------------------------------------
void MenuIstData(void)
{                             //1234567890123456
  static unsigned char index = 0, blink = 0, count = 0;
  
  if (MenuPolling)
  {  lcd_gotoxy(2,0); lcd_putsf("             %");
     MenuPolling = 0;
   }                  

//#define LCD_XY(x,y) {_lcd_write_data(0x80 + x + (y * (0xc0-0x80)));}
   
  if (RefreshDisplay)
  {
   lcd_gotoxy(0,0); lcd_putsf(&Tag[AktuellZeit / 1440] [0]);
   
   lcd_gotoxy(3 ,0); lcd_int((AktuellZeit % 1440) / 60, 2 , LEADING_ZERO); 
   if (blink) { blink = 0; lcd_putchar(':'); } 
    else { blink = 1; lcd_putchar(' '); } 
     
   lcd_int((AktuellZeit % 1440) % 60, 2 ,LEADING_ZERO);
   lcd_gotoxy(0,1); 
   if (count++ & 0x08) 
   { lcd_putsf("Ta=");  lcd_int(AussenTemperatur/10, -2, 0); 
   }                                                         
   else
   { lcd_putsf("Ts=");  lcd_int(SollKesselTemperatur/10, -2, 0); 
   }                                                         
   
   lcd_putchar(GRAD_ZEICHEN); lcd_putchar(' ');
   lcd_putsf("Tk=");  lcd_int(KesselTemperatur/10, -2, 0); lcd_putchar(GRAD_ZEICHEN); lcd_putchar('C');
 
   lcd_gotoxy(12,0); lcd_int(((long) CountHeizzeit * 100) / CountZeit, 3, 0);
 
 
   lcd_gotoxy(10,0);
   switch (BetriebsMode)
     {
       case TAG:   
                       lcd_putchar(BetriebsMode+1);
          break;
       case NACHT:  
                       lcd_putchar(BetriebsMode+1);
          break;
       case ABWESEND:
                       lcd_putchar(BetriebsMode+1);
          break;
       case DEAKTIV: 
                       lcd_putchar(' ');
          break;
     }
   RefreshDisplay = 0; 
  } 

#ifdef EXTRA_MENU  
    if (index == 0 && (Taste == UP || Taste == DOWN)) { MenuCount = 255; MenuPolling =1;}
#endif
  
    switch(index)
     { case 0: break;
     
       case 1:        
                lcd_gotoxy(2,0); lcd_putchar('\x7e');  
                if (Taste == UP) AktuellZeit += 60; 
                 else if (Taste == DOWN) if (AktuellZeit > 60) AktuellZeit -= 60; 
            break;
       case 2: 
               lcd_gotoxy(2,0); lcd_putchar(' ');  
               lcd_gotoxy(8,0); lcd_putchar('\x7f');
               if (Taste == UP) AktuellZeit ++; 
                else if (Taste == DOWN) if (AktuellZeit > 0) AktuellZeit --;
            break;
       case 3: 
                 lcd_gotoxy(8,0); lcd_putchar(' ');
                 lcd_gotoxy(2,0); lcd_putchar('\x7f');
                 if (Taste == UP) if (AktuellZeit < 6*1440)  AktuellZeit += 1440; else AktuellZeit %= 1440;
                   else if (Taste == DOWN) if (AktuellZeit > 1440) AktuellZeit -=1440; else AktuellZeit += 6*1440;
            break;
     }
     
     if (Taste == OKAY) 
     {
        if (index < 3) { index++; }
        else 
        { 
          index = 0;      
          MenuPolling = 1;
        }
     }                     
    
}
// ---------------------------------------------------------------------------------------
unsigned char EditTimer(unsigned char pos)
{
  static unsigned int tmp_wert, zeit, tagZeit;
  static unsigned char tmp_mode, index = 0;
  
  pos += SchichtOffset;
  
 if(MenuPolling)
 {
    MenuPolling = 0;
    lcd_gotoxy(2,0); lcd_putsf("              ");

   lcd_gotoxy(0,0);
   if (Timer[pos].Zeit & WERKTAG) lcd_putsf("WT");
     else lcd_putsf(&Tag[Timer[pos].Zeit % (ZEIT_MASKE) / 1440] [0]);
    
  
   lcd_gotoxy(0,1); lcd_putsf("Mode   ");
   zeit = Timer[pos].Zeit & ~(0x4000+0x8000);
   tmp_mode = Timer[pos].Mode;
   tmp_wert = zeit;                            
   tagZeit = (zeit / 1440)*1440; 
   if (Timer[pos].Zeit & WERKTAG) tagZeit = WERKTAG;
  }
   
     lcd_gotoxy(7 ,0); lcd_int((tmp_wert % 1440) / 60, 2, LEADING_ZERO); lcd_putchar(':');
                       lcd_int((tmp_wert % 1440) % 60, 2, LEADING_ZERO);

     lcd_gotoxy(7,1);
     switch(tmp_mode)
     {
       case TAG:      lcd_putsf("Tag      "); 
                       lcd_gotoxy(14,0);
                       lcd_putchar(tmp_mode+1);
          break;
       case NACHT:    lcd_putsf("Nacht    "); 
                       lcd_gotoxy(14,0);
                       lcd_putchar(tmp_mode+1);
          break;
       case ABWESEND: lcd_putsf("Abwesend "); 
                       lcd_gotoxy(14,0);
                       lcd_putchar(tmp_mode+1);
          break;
       case DEAKTIV: lcd_putsf("DEAKTIV  "); 
                       lcd_gotoxy(14,0);
                       lcd_putchar('-');
          break;
     }

     switch(index)
     { case 0:
                lcd_gotoxy(6,0); lcd_putchar('\x7e');  
                if (Taste == UP) tmp_wert += 60; 
                 else if (Taste == DOWN) tmp_wert -= 60;
            break;
       case 1: 
               lcd_gotoxy(6,0); lcd_putchar(' ');  
               lcd_gotoxy(12,0); lcd_putchar('\x7f');
               if (Taste == UP) tmp_wert ++; 
                else if (Taste == DOWN) tmp_wert --;
            break;
       case 2: 
                 lcd_gotoxy(12,0); lcd_putchar(' ');  
                 lcd_gotoxy(6,1); lcd_putchar('\x7e');
                 if (Taste == UP) if (tmp_mode < 3) tmp_mode++; else tmp_mode = 0;
                else if (Taste == DOWN) if (tmp_mode > 0) tmp_mode--; else tmp_mode = 3;
            break;
     }
     
     if (Taste == OKAY) 
     {
        if (index < 2) { index++;  }
        else 
        { 
          Timer[pos].Mode = tmp_mode;
          Timer[pos].Zeit = (tmp_wert % 1440) + tagZeit;
          index = 0;      
          MenuPolling = 1;
          return(1);
        }
     }                     
   return(0);
}

// ---------------------------------------------------------------------------------------
void ZeigTimer(char x, char y, unsigned char pos)
{
  char sym[] = "\1\2\3";
  unsigned int zeit;  
  
  pos += SchichtOffset;
  
  zeit = Timer[pos].Zeit & ~(0x4000+0x8000);
  
  lcd_gotoxy(x,y);
     
  if (Timer[pos].Mode == DEAKTIV) lcd_putsf(" --:--"); 
   else 
   {  lcd_putchar(sym[Timer[pos].Mode]);
      lcd_int((zeit % 1440) / 60,2, LEADING_ZERO); lcd_putchar(':'); lcd_int((zeit % 1440) % 60, 2, LEADING_ZERO); 
   }
}
// ---------------------------------------------------------------------------------------
void TimerMenu(void)
{                 
  static unsigned char pos=0, index=0, edit = 0,clearlcd = 1;

  if (MenuPolling) 
   { //pos=0; index=0; 
     edit = 0; MenuPolling = 0; clearlcd = 1;
   };
  
  if (edit == 2)  // EDIT TIMER
  {
    if(EditTimer(pos+index)) edit = 0;   
  
  }
  else       // Select Timer
  {
  if (clearlcd)
  {  
     lcd_gotoxy(0,0);
     if (Timer[pos].Zeit & WERKTAG) lcd_putsf("WT");                           
      else lcd_putsf(&Tag[Timer[pos].Zeit % (ZEIT_MASKE) / 1440] [0]);
      
     lcd_gotoxy(0,1); lcd_int((SchichtOffset/32)+1, 1, 0);
     
     ZeigTimer(3 ,0, pos);
     ZeigTimer(10,0, pos+1);
     ZeigTimer(3 ,1, pos+2);
     ZeigTimer(10,1, pos+3);
     clearlcd = 0;   
     lcd_gotoxy(2,0); lcd_putchar(' ');   lcd_gotoxy(9,0); lcd_putchar(' '); 
     lcd_gotoxy(1,1); lcd_putchar(' ');   lcd_putchar(' ');  lcd_gotoxy(9,1); lcd_putchar(' '); 
  }                  
  
  clearlcd = 1; 
  
  if (edit == 0)
  { if(Taste == UP) if (pos < 28) pos += 4; 
     else
        { pos = 0; 
          #ifdef EXTRA_MENU
          MenuCount = 255; 
          #endif
          }
    else if(Taste == DOWN) if (pos > 0) pos -= 4; else pos = 28;
    else clearlcd = 0;    
  }
  else
  {
    if (Taste==UP) if(index < 3) index++; else index = 0;
    else if (Taste==DOWN) if (index > 0) index--; else index = 3; 
    else clearlcd = 0;
    switch(index)
    {
      case 0: lcd_gotoxy(2,0); break;
      case 1: lcd_gotoxy(9,0); break;
      case 2: lcd_gotoxy(2,1); break;
      case 3: lcd_gotoxy(9,1); break;
      
    }
    lcd_putchar('\x7e'); 
  }
  
  if (Taste == OKAY) 
   { if (edit == 0) edit = 1;
      else if (edit == 1) 
      {  edit = 2;               
         MenuPolling = 1;
         Taste = NICHTS;
         EditTimer(pos+index);
      }   
      clearlcd = 1;        
       
   }
 }   
}

// ---------------------------------------------------------------------------------------
void SchichtMenu(void)
{                          //1234567890123456
  if (MenuPolling)
  { // lcd_gotoxy(0,0); lcd_putsf("Timer - Gruppe   ");
     lcd_gotoxy(0,1); lcd_putsf("setzen");
     MenuPolling = 0;
  }                  
                                            
  lcd_gotoxy(10,1); lcd_int(SchichtOffset / 32 +1, 1, 0);
  
  if ((Taste == UP) || (Taste == DOWN)) 
  {  if (!SchichtOffset) SchichtOffset = 32; else SchichtOffset = 0;
  }
  
#ifdef EXTRA_MENU
   if (Taste == OKAY) { MenuCount = 255; };
 #endif    
}

// ---------------------------------------------------------------------------------------
void TemperaturMenu(void)
{                          //1234567890123456
  static unsigned char kennlinie = 0, index =0;
  
  if (MenuPolling)
  {                            //1234567890123456
     lcd_gotoxy(0,0); lcd_putsf("  \x4   \xdf bei    \xdf");
     lcd_gotoxy(0,1); lcd_putsf("  \x5   \xdf bei    \xdf");
     MenuPolling = 0;
  }                  
  
  lcd_gotoxy(0,0); lcd_putchar(kennlinie+1);
  
  lcd_gotoxy(4,0);  lcd_int( MaxKesselTemperatur[kennlinie]/10, 2, 0); 
  lcd_gotoxy(4,1);  lcd_int( MinKesselTemperatur[kennlinie]/10, 2, 0); 
  lcd_gotoxy(12,0); lcd_int( KennlinieTemperaturOben[kennlinie]/10, -2, 0); 
  lcd_gotoxy(12,1); lcd_int( KennlinieTemperaturUnten[kennlinie]/10, -2, 0); 
 
  switch (index)
  {
    case 0:
            lcd_gotoxy(11,1); lcd_putchar(' ');
            if (Taste == UP)   if (kennlinie < 2) kennlinie++; else 
            { kennlinie=0;
              #ifdef EXTRA_MENU
                MenuCount = 255; 
              #endif    
            }
            
            if (Taste == DOWN) if (kennlinie > 0) kennlinie--; else kennlinie=2;
        break;
  
    case 1:
               lcd_gotoxy(7,0); lcd_putchar('\x7f');
            if (Taste == UP)   MaxKesselTemperatur[kennlinie]+=10;
            if (Taste == DOWN) MaxKesselTemperatur[kennlinie]-=10;
        break;
    case 2:
               lcd_gotoxy(7,0); lcd_putchar(' ');  
               lcd_gotoxy(11,0); lcd_putchar('\x7e');
            if (Taste == UP)   KennlinieTemperaturOben[kennlinie]+=10;
            if (Taste == DOWN) KennlinieTemperaturOben[kennlinie]-=10;
        break;
    case 3:
               lcd_gotoxy(11,0); lcd_putchar(' ');  
               lcd_gotoxy(7,1); lcd_putchar('\x7f');
            if (Taste == UP)   MinKesselTemperatur[kennlinie]+=10;
            if (Taste == DOWN) MinKesselTemperatur[kennlinie]-=10;
        break;

    case 4:
               lcd_gotoxy(7,1); lcd_putchar(' ');  
               lcd_gotoxy(11,1); lcd_putchar('\x7e');
            if (Taste == UP)   KennlinieTemperaturUnten[kennlinie]+=10;
            if (Taste == DOWN) KennlinieTemperaturUnten[kennlinie]-=10;
        break;
  }
  
  if (Taste == OKAY) if (index <4) index++; else index = 0;

}
// ---------------------------------------------------------------------------------------
void PumpenMenu(void)
{                          
  static unsigned char tmp_mode=0;
  
  if (MenuPolling)                          
  {  
     MenuPolling = 0;
     
     lcd_gotoxy(0,0); lcd_putsf("Pumpe aus       ");
     lcd_gotoxy(0,1); lcd_putsf("      Tk <    \xdfC");
     
     switch(tmp_mode)
     { case NACHT:                       //1234567890123456
                lcd_gotoxy(12,0); lcd_putchar(tmp_mode+1);
                lcd_gotoxy(0,1); lcd_putsf("Nacht");
        break; 
     
       case TAG:
                lcd_gotoxy(12,0); lcd_putchar(tmp_mode+1);
                lcd_gotoxy(0,1); lcd_putsf("Tag");
        break; 
     
      case ABWESEND:
                lcd_gotoxy(12,0); lcd_putchar(tmp_mode+1);
                lcd_gotoxy(0,1); lcd_putsf("Abw.");
        break; 
     }
  }                  
                                            
  lcd_gotoxy(11,1);
  lcd_int( PumpeEinTemperatur[tmp_mode]/10, -2, 0);
 
  switch (tmp_mode)
  {
    case 0:
    case 1:
    case 2:
            if (Taste == UP)   PumpeEinTemperatur[tmp_mode]+=10;
            if (Taste == DOWN) PumpeEinTemperatur[tmp_mode]-=10;
        break;
  
   }

  if (Taste == OKAY) 
   {  if (tmp_mode < 2) tmp_mode++; else 
      { tmp_mode = 0;
        #ifdef EXTRA_MENU
        MenuCount = 255; 
        #endif    
      }
      MenuPolling = 1;
   }   
}

// ---------------------------------------------------------------------------------------
void HistoryMenu(void)
{                          
  static unsigned char index=0, tmp_verb[5], tmp_temp[5];
  unsigned char i; 
  
  if (MenuPolling)                          
  {  
     MenuPolling = 0;
     index = ZeigerRingspeicher-1;          
    
     for (i = 0; i < 5; i++)
     {
       if (index >= ANZ_LOGGING) index = ANZ_LOGGING-1;
       tmp_verb[i] = VerbrauchsRingSpeicher[index];
       tmp_temp[i] = TemperaturRingSpeicher[index];
       index--;     
     }  
  }
  if (RefreshDisplay)
  {
     lcd_gotoxy(0,0);
     for (i=0;i<5;i++) if(tmp_verb[i]<=100) lcd_int( tmp_verb[i], 3, 0);  else lcd_putsf(" - ");      
     lcd_putchar('%');
     lcd_gotoxy(0,1);                                         
     for (i=0;i<5;i++) if(tmp_temp[i]<=100) lcd_int( tmp_temp[i], -2, 0); else lcd_putsf(" - ");    
     lcd_putchar('\xdf');
  }
  
  if ((Taste == UP) ||  (Taste == DOWN) ||  (Taste == OKAY)) 
   {  
     MenuCount = 255; 
     MenuPolling = 1;   
   }   
}

// ---------------------------------------------------------------------------------------
#ifdef EXTRA_MENU
void MenuAuswahl(void)
{
  static unsigned char index = 0;
  
   if (RefreshDisplay)
   {  
     lcd_gotoxy(0,1); lcd_putsf("                 ");
     lcd_gotoxy(0,0); 
     switch(index) 
     { case 0: 
                lcd_putsf("Status           ");      
              break;
       
       case 1: 
                lcd_putsf("Timer - Setup    ");
              break;
      
       case 2: 
                lcd_putsf("Schalt - Gruppe  ");
              break;
                
       case 3: 
                lcd_putsf("Kennlinien       ");
              break;
      
       case 4: 
                lcd_putsf("Wasserpumpe      ");
              break;

       case 5: 
                lcd_putsf("Verbrauch \xf5ber   ");
                lcd_gotoxy(0,1); 
                lcd_putsf("5 Tage");

              break;
     }         
   }                                        
  
  if (Taste == UP) if (index < 5) index++; else index = 0;
  else if (Taste == DOWN) if (index > 0) index--; else index = 5;
  
  if (Taste == OKAY)
  {
    MenuCount = index;
    MenuPolling = 1;
  } 
  
}
#endif

// ---------------------------------------------------------------------------------------
void GetRegelpunkt(void)
{  
 signed long soll,dK,dT;
 if(AussenTemperatur < KennlinieTemperaturOben[BetriebsMode]) 
  {
   SollKesselTemperatur = MaxKesselTemperatur[BetriebsMode];
  }
 else  
 if(AussenTemperatur > KennlinieTemperaturUnten[BetriebsMode]) 
  {
   SollKesselTemperatur = MinKesselTemperatur[BetriebsMode];
  }
 else
  {
   dK = MaxKesselTemperatur[BetriebsMode] - MinKesselTemperatur[BetriebsMode];
   dT = KennlinieTemperaturUnten[BetriebsMode] - KennlinieTemperaturOben[BetriebsMode];
   soll = MaxKesselTemperatur[BetriebsMode] - ((AussenTemperatur - KennlinieTemperaturOben[BetriebsMode]) * dK) / dT;
   SollKesselTemperatur = soll; 
  }
}
#define AUS 1
#define EIN 0

// ---------------------------------------------------------------------------------------
void SteuerRelais(void)
{
static unsigned char BetriebsMode_alt = 0, warte;
 if(KesselTemperatur > SollKesselTemperatur + KesselHysterese) RELAIS_KESSEL = AUS; else // aus
 if((KesselTemperatur) < SollKesselTemperatur)                 RELAIS_KESSEL = EIN;      // ein
 
 if(BetriebsMode_alt == BetriebsMode)
  {
   if ((KesselTemperatur < PumpeEinTemperatur[BetriebsMode]) && 
    (SollKesselTemperatur < PumpeEinTemperatur[BetriebsMode]))           RELAIS_PUMPE = AUS; // weniger als xx  im Kessel AUS                                                            
   else 
   if((SollKesselTemperatur > PumpeEinTemperatur[BetriebsMode] + 10) && 
     (KesselTemperatur > PumpeEinTemperatur[BetriebsMode]))              RELAIS_PUMPE = EIN;
  }   
  else // Der Betriebmode hat sich gendert --> wir warten etwas
  {
   if(--warte == 0) BetriebsMode_alt = BetriebsMode;
  }
}

// ---------------------------------------------------------------------------------------
void Menu(void)
{                   
 // if (Taste == MENU) { MenuCount++; MenuPolling = 1; }
  
  if (Taste == MODE) 
  { 
    if (BetriebsMode < 2) BetriebsMode++; else BetriebsMode = 0; 
    MenuPolling = 1;
    LCDInit();
  }
  
  switch (MenuCount)
  {
    case 0:  MenuIstData();
      break;
  
    case 1:  TimerMenu();
      break;

    case 2:  SchichtMenu();
      break;

    case 3:  TemperaturMenu();
      break;
       
    case 4:  PumpenMenu();
      break;             
      
    case 5: HistoryMenu();
      break;   
    default:
    #ifdef EXTRA_MENU   
       MenuAuswahl();
    #else
      MenuCount = 0;
    #endif     
     break;
  
  }
  
}
// ---------------------------------------------------------------------------------------
void InitEEPROM(void)
{
  unsigned char i;
  
  if (CheckEEpromData != EEPROM_KOMPATIBILITAET)
  {
   lcd_clear();
   lcd_putsf("INIT EEPROM");
   lcd_gotoxy(0,1);
   for (i = 0; i < 64; i++)
   {
    delay_ms(15);
    switch (i % 4)
    { case 0:  Timer[i].Zeit = ZEIT(7,00);  Timer[i].Mode = TAG;      break;
      case 1:  Timer[i].Zeit = ZEIT(9,00);  Timer[i].Mode = ABWESEND; break;
      case 2:  Timer[i].Zeit = ZEIT(17,00); Timer[i].Mode = DEAKTIV;  break;
      case 3:  Timer[i].Zeit = ZEIT(22,00); Timer[i].Mode = NACHT;    lcd_putchar('.'); break;
    }   
    
    if ((i%32)<4) Timer[i].Zeit |= WERKTAG;
    else
    if ((i%32)<12) Timer[i].Zeit += (((i%32-4)/4)+5) * 1440;
    else     
     { Timer[i].Zeit += (((i%32-12)/4)) * 1440;
       Timer[i].Mode = DEAKTIV;
     }
   }

 
    MaxKesselTemperatur[TAG] = 750;
    MinKesselTemperatur[TAG] = 300;
    KennlinieTemperaturOben[TAG] = -100;
    KennlinieTemperaturUnten[TAG] = 200;
    PumpeEinTemperatur[TAG] = 390;

    MaxKesselTemperatur[NACHT] = 500;
    MinKesselTemperatur[NACHT] = 300;
    KennlinieTemperaturOben[NACHT] = -150;
    KennlinieTemperaturUnten[NACHT] = 200;
    PumpeEinTemperatur[NACHT] = 450;
 
    MaxKesselTemperatur[ABWESEND] = 600;
    MinKesselTemperatur[ABWESEND] = 300;
    KennlinieTemperaturOben[ABWESEND] = -150;
    KennlinieTemperaturUnten[ABWESEND] = 200;
    PumpeEinTemperatur[ABWESEND] = 440;          
    
    SchichtOffset = 0;   
     
   for(i = 0 ; i < ANZ_LOGGING ;i++)
   { VerbrauchsRingSpeicher[i] = 0xFF;   
     TemperaturRingSpeicher[i] = 0xFF;
   }
    ZeigerRingspeicher = 0;    
    
    CheckEEpromData = EEPROM_KOMPATIBILITAET;

  }
}

// ---------------------------------------------------------------------------------------
void SpeicherVerbrauch(void)
{
  VerbrauchsRingSpeicher[ZeigerRingspeicher] = ((long) CountHeizzeit * 100) / CountZeit;
  TemperaturRingSpeicher[ZeigerRingspeicher] = DurchschnittsTemperatur / 1440;
  DurchschnittsTemperatur = 0;
  CountHeizzeit = 0;
  CountZeit = 0;
  if (ZeigerRingspeicher++ >= ANZ_LOGGING) ZeigerRingspeicher = 0; 
}

// ---------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------
void main(void)
{
// Declare your local variables here
  static unsigned char tmp_mode;      
// Declare your local variables here

// Input/Output Ports initialization
// Port A initialization
// Func0=In Func1=In Func2=In Func3=In Func4=In Func5=In Func6=In Func7=In 
// State0=T State1=T State2=T State3=T State4=T State5=T State6=T State7=T 
PORTA=0x00;
DDRA=0x00;

// Port B initialization
// Func0=In Func1=In Func2=In Func3=In Func4=In Func5=In Func6=In Func7=In 
// State0=T State1=T State2=T State3=T State4=T State5=T State6=T State7=T 
PORTB=0x00;
DDRB=0x1f;

// Port C initialization
// Func0=In Func1=In Func2=In Func3=In Func4=In Func5=In Func6=In Func7=In 
// State0=T State1=T State2=T State3=T State4=T State5=T State6=T State7=T 
PORTC=0x00;
DDRC=0x00;

// Port D initialization
// Func0=In Func1=In Func2=In Func3=In Func4=In Func5=In Func6=In Func7=In 
// State0=T State1=T State2=T State3=T State4=T State5=T State6=T State7=T 
PORTD=0x00;
DDRD=0x00;

// Timer/Counter 0 initialization
// Clock source: System Clock
// Clock value: Timer 0 Stopped
// Mode: Normal top=FFh
// OC0 output: Disconnected
TCCR0=0x02;
TCNT0=0x64;
OCR0=0x00;

// Timer/Counter 1 initialization
// Clock source: System Clock
// Clock value: Timer 1 Stopped
// Mode: Normal top=FFFFh
// OC1A output: Discon.
// OC1B output: Discon.
// Noise Canceler: Off
// Input Capture on Falling Edge
TCCR1A=0x00;
TCCR1B=0x00;
TCNT1H=0x00;
TCNT1L=0x00;
OCR1AH=0x00;
OCR1AL=0x00;
OCR1BH=0x00;
OCR1BL=0x00;

// Timer/Counter 2 initialization
// Clock source: System Clock
// Clock value: 125,000 kHz
// Mode: Normal top=FFh
// OC2 output: Disconnected
ASSR=0x00;
TCCR2=0x0c;
TCNT2=0x00;
OCR2= 124;         //  1ms         

// External Interrupt(s) initialization
// INT0: Off
// INT1: On
// INT1 Mode: Falling Edge
// INT2: Off
#ifdef INCR_GEBER
GICR|=0x80;
MCUCR=0x08;
MCUCSR=0x00;
GIFR=0x80;
#endif

// Timer(s)/Counter(s) Interrupt(s) initialization
TIMSK=0x80;

// USART initialization
// Communication Parameters: 8 Data, 1 Stop, No Parity
// USART Receiver: On
// USART Transmitter: On
// USART Mode: Asynchronous
// USART Baud rate: 9600
UCSRA=0x00;
UCSRB=0xD8;
UCSRC=0x86;
UBRRH=0x00;
UBRRL=0x33;

// Analog Comparator initialization
// Analog Comparator: Off
// Analog Comparator Input Capture by Timer/Counter 1: Off
// Analog Comparator Output: Off
ACSR=0x80;
SFIOR=0x00;

// ADC initialization
// ADC Clock frequency: 125,000 kHz
// ADC Voltage Reference: AREF pin
//ADMUX = ADC_VREF_TYPE;
ADCSRA=0x86;


 LCDInit();
 
 if (TASTE4) CheckEEpromData = 0xFF; // EEPROM neu init, wenn OKAY-Taste gedrckt
         //1234567890123456
 lcd_gotoxy(0,0);
 lcd_putsf("I. Busker");
 lcd_gotoxy(0,1);
 lcd_putsf("H. Buss");
 
 delay_ms(1500);
 InitEEPROM();
 
 GetTemperatur(1); 

// Global enable interrupts
#asm("sei")

while (1)
      {
           delay_ms(40);
           tmp_mode = ErmittleBetriebsMode(AktuellZeit);
           if (tmp_mode != DEAKTIV) BetriebsMode = tmp_mode;
           
           Menu();
           GetTemperatur(0);
           GetRegelpunkt();
           SteuerRelais();

           Tastatur();              

           if (BetriebsMode == TAG)      RELAIS_TAG = 1;   else RELAIS_TAG = 0;
           if (BetriebsMode == NACHT)    LED_NACHT = 1;    else LED_NACHT = 0;
           if (BetriebsMode == ABWESEND) LED_ABWESEND = 1; else LED_ABWESEND = 0;
 
           if(PollingTag) 
            { SpeicherVerbrauch();
              PollingTag = 0;
            }
            
           if (SendeDisplay)
            {                          
              unsigned char i;
              
              Menu();
              SendeDisplay = 0;
              Output = SERIELL;
#ifndef NO_LCD
              lcd_write_byte(0xC0+16,(PINB & 0x1F) + 0x30);
              for (i = 0x80; i < 0x80+16; i++)  putchar(lcd_read_byte(i)); 
              putchar('\r');
              for (i = 0xC0; i < 0xC0+17; i++)  putchar(lcd_read_byte(i)); 
              putchar('\r');
#else
              for (i = 0;  i < 16; i++)  putchar(LCD_Puffer[i]); 
              LCD_Puffer[32] = (PINB & 0x1F) + 0x30;
              putchar('\r');
              for (i = 16; i < 33; i++)  putchar(LCD_Puffer[i]); 
              putchar('\r');
#endif
            }
            
           if (SendeDaten)
            {                          
              unsigned char tmp_speicher;
              
              tmp_speicher = ZeigerRingspeicher-1;
              
              SendeDaten = 0;
              
              Output = SERIELL;
              putsf("[%],[C]");
              while(tmp_speicher != ZeigerRingspeicher)
              {
                if (tmp_speicher >= ANZ_LOGGING) tmp_speicher = ANZ_LOGGING-1;
                if (tmp_speicher == ZeigerRingspeicher) break;
             
                if (VerbrauchsRingSpeicher[tmp_speicher]<=100) 
                {
                  Seriell_int(VerbrauchsRingSpeicher[tmp_speicher], 3, 0); putchar(',');
                  Seriell_int(TemperaturRingSpeicher[tmp_speicher], -2 , 0); putchar('\r');
                // Seriell_int(tmp_speicher % 100, 3, 0); putchar(',');
                 // Seriell_int((char) (100- tmp_speicher), -3 , 0); putchar('\r');
                }
                tmp_speicher--; 
              } 
              
            }
      };
}
